
#define DBG_GUI		0

#define VUC_DBG_ON	0
#define VUC_DBG_PROTO	1
#define VUC_DBG_IN	1
#define VUC_DBG_OUT	1

#define SHOW_TIMESTAMP	0


#define VUC_IS_PC	1

#include "config.h"
#include "demo.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#ifdef _WIN32

#include <windows.h>

#else /* _WIN32 */

#include <unistd.h>

#endif /* _WIN32 */

#include "anygtk.h"
#include "my-pc.h"
#include "vuc.h"

#if (COM_SER)
#  include "eia-232.h"
#endif
#if (COM_UNIX)
#  include "unix.h"
#endif


guint vuc_proto_opts= 0xffff;		// default: use all compiled-in options
#define VUC_OPT_CHK_PACKET	(vuc_proto_opts & VUC_OPT_FLG_PACKET)
#define VUC_OPT_CHK_PKT_CHK	(vuc_proto_opts & VUC_OPT_FLG_PKT_CHK)
#define VUC_OPT_CHK_PKT_SEQ	(vuc_proto_opts & VUC_OPT_FLG_PKT_SEQ)

gchar * glade= "demo.glade";

GTimeVal start;

gint vuc_err_cnt[256];
gint vuc_alive_cnt= 0;
#if (VUC_PACKET)
gint vuc_in_chk_cnt= 0;
gint vuc_out_chk_cnt= 0;
gint vuc_line_err_cnt= 0;
#endif
gint vuc_proto_err_cnt= 0;

#if (VUC_PKT_SEQ)
gint vuc_is_warm= 0;			// 0=Let uC initialize vuc_in_pkt_no
u8 vuc_in_pkt_no= 0;			// Expected packet no
u8 vuc_in_pkt_hs_no= 0;			// Last packet no
u8 vuc_out_pkt_no;
#else
#  define vuc_in_pkt_no		0
#  define vuc_out_pkt_no	0
#endif
#if (VUC_PACKET)
u8 vuc_in_pkt_hs_code= VUC_META_PKT_BAD;// Last handshake sent
u8 vuc_out_pkt_wait= 0;			// Packet sent - waiting for handshake
#endif

gint scope_diag_ok= 0;
gint scope_diag_err= 0;


gboolean vuc_io_in (GIOChannel * source, GIOCondition condition, gpointer data);
gboolean vuc_io_timeout (gpointer data);
gboolean backgr_func (gpointer data);

GAsyncQueue * console_q;
GAsyncQueue * send_q;
//GtkStatusbar * statusbar1;
GtkEntry * gui_alive_cnt= NULL;
#if (VUC_PACKET)
GtkEntry * gui_in_chk_cnt= NULL;
GtkEntry * gui_out_chk_cnt= NULL;
GtkEntry * gui_line_err_cnt= NULL;
#endif
GtkEntry * gui_proto_err_cnt= NULL;
GtkEntry * gui_status_msg= NULL;

guint status_msg_id;


i8 scope_in (u8 * data);
void vuc_in_pkt_start (void);
void vuc_in_send_hs (i8 rep);



#if (VUC_DBG_ON)
// debugging generally on
#  define dbgf(...)	msgf(__VA_ARGS__)
#else
#  define dbgf(...)	
#endif

#if (VUC_DBG_ON)
// serious stuff
#  define dbgf0(...)	dbgf(__VA_ARGS__)
#else
#  define dbgf0(...)	
#endif

#if (VUC_DBG_IN)
// basic i/o
#  define dbgf1(...)	dbgf(__VA_ARGS__)
#else
#  define dbgf1(...)	
#endif

#if (VUC_DBG_OUT)
// basic i/o
#  define dbgf2(...)	dbgf(__VA_ARGS__)
#else
#  define dbgf2(...)	
#endif

#if (VUC_DBG_PROTO)
// protocol
#  define dbgf3(...)	dbgf(__VA_ARGS__)
#else
#  define dbgf3(...)	
#endif

#if (DBG_GUI)
#  define dbgf4(...)	dbgf(__VA_ARGS__)
#else
#  define dbgf4(...)	
#endif


/*****************************************************************************/

// returns no of milliseconds this program is running (assuming it was started
// when tv_usec was 0. If not, there's a constant offset below 1 second)
i32 my_time ()
{
  GTimeVal now;

  g_get_current_time(&now);
  return (now.tv_sec-start.tv_sec)*1000 + (now.tv_usec/1000);
}

int msgf (char * fmt, ...)
{
  va_list args;

#if (SHOW_TIMESTAMP)
  i32 now= my_time();
  printf("%d.%03d ", now/1000, now%1000);
#endif

  va_start(args, fmt);
  vprintf(fmt, args);
  printf("\n");
  va_end(args);
}

int errf (char * fmt, ...)
{
  va_list args;

#if (SHOW_TIMESTAMP)
  i32 now= my_time();
  printf("%d.%03d ERROR! ", now/1000, now%1000);
#endif

  va_start(args, fmt);
  vprintf(fmt, args);
  printf("\n");
  va_end(args);
}


#if (COM_SER)

EIA232_HANDLE ser_handle;

gint if_ser_init (gchar * com_name, gint baudrate)
{
  if (EIA232_open(com_name, baudrate, &ser_handle)<0) {
    return 0;
  }
  g_timeout_add(5, vuc_io_timeout, NULL);
  return 1;
}

gsize if_ser_receive (gpointer buffer, gint buf_size)
{
  return EIA232_poll(ser_handle, buffer, buf_size);
}

gint if_ser_send_8 (u8 data)
{
  EIA232_send_byte(ser_handle, data);
  return 1;
}

#endif /* COM_SER */


#if (COM_UNIX)

gsize if_unix_receive (gpointer buffer, gint buf_size)
{
  return unix_gets(buffer, buf_size);
}

gint if_unix_send_8 (u8 data)
{
  unix_putc(data);
  return 1;
}

#endif /* COM_UNIX */


typedef gsize (* receive_f)(gpointer buffer, gint buf_size);
typedef gint (* send_8_f)(u8 data);

receive_f if_receive;
send_8_f if_send_8;

#if (COM_UNIX)
#  define COM_UNIX_OPT_STR	", 'unix' for unix domain socket"
#else
#  define COM_UNIX_OPT_STR	""
#endif

gint if_init (int argc, char * argv[])
{
  gchar * if_type= NULL;
  gchar ** remain= NULL;
  gboolean show_config= FALSE;
  GOptionEntry entries[] = {
    { "gui-def", 'g', 0, G_OPTION_ARG_FILENAME, &glade, "GUI definition file", "FILE" },
    { "if-type", 't', 0, G_OPTION_ARG_STRING, &if_type, "Interface type: 'ser' for RS232 (default)" COM_UNIX_OPT_STR, "TYPE" },
    { "show-cfg", 's', 0, G_OPTION_ARG_NONE, &show_config, "Show configuration", NULL },
    { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_STRING_ARRAY, &remain, NULL, NULL },
    { NULL }
  };
  GError * error= NULL;
  GOptionContext * context;

  context= g_option_context_new("[ [--] interface parameters]");
  g_option_context_add_main_entries(context, entries, NULL);
  g_option_context_add_group(context, gtk_get_option_group(TRUE));
  if (!g_option_context_parse (context, &argc, &argv, &error)) {
    msgf("option parsing failed: %s\n", error->message);
    exit(1);
  }
  g_option_context_free(context);

  if (show_config) {
    msgf("Configuration:\n");
    msgf("Option	 compiled in  active\n");
    msgf("----------------------------\n");
    msgf("VUC_PACKET	%d	%d\n", VUC_PACKET, VUC_OPT_CHK_PACKET ? 1 : 0);
    msgf("VUC_PKT_CHK	%d	%d\n", VUC_PKT_CHK, VUC_OPT_CHK_PKT_CHK ? 1 : 0);
    msgf("VUC_PKT_SEQ	%d	%d\n", VUC_PKT_SEQ, VUC_OPT_CHK_PKT_SEQ ? 1 : 0);
    exit(0);
  }

  if (!if_type || strcasecmp(if_type, "ser")==0) {
    gint bd= 0;
    if (!remain || !remain[0]) {
      msgf("Need serial port name.\n");
      exit(1);
    }
    if (remain[1])
      bd= atoi(remain[1]);
    if (!bd)
      bd= BAUDRATE;
    msgf("Opening serial port '%s' with %d baud.", remain[0], bd);
    if (if_ser_init(remain[0], bd)) {
      if_receive= if_ser_receive;
      if_send_8= if_ser_send_8;
      return 1;
    } else
      exit(3);
  }

#if (COM_UNIX)
  if (strcasecmp(if_type, "unix")==0) {
    gchar * socket= remain[0] ? remain[0] : UNIX_SERVER_PATH;
    msgf("Opening socket '%s'.\n", socket);
    if (unix_init(socket)) {
      if_receive= if_unix_receive;
      if_send_8= if_unix_send_8;
      g_timeout_add(5, vuc_io_timeout, NULL);
      return 1;
    } else {
      exit(3);
//      return 0;
    }
  }
#endif

  msgf("Bad port type.\n");
  exit(2);
}

gint if_send_16 (u16 data)
{
  u16 ve_data= VUC_U16_TO_VE(data);

  if_send_8(ve_data&0xff);
  if_send_8(ve_data>>8);
  return 1;
}


/*** VUC *********************************************************************/

#if (VUC_DBG_ON)

void vuc_dbg_beg (u8 len)
{ }

void vuc_dbg_end (void)
{ dbgf(""); }

void vuc_dbgs (void * str)
{ dbgf(str); }

void vuc_dbgn4 (u8 n)
{ dbgf("%1x", n&0x0f); }

void vuc_dbgn8 (u8 n)
{ dbgf("%02x", n); }

#else

#  define vuc_dbg_beg(len) {}
#  define vuc_dbg_end() {}
#  define vuc_dbgs(str) {}
#  define vuc_dbgn4(n) {}
#  define vuc_dbgn8(n) {}

#endif


typedef struct {
  gint pkt, start, end;
  u8 buf[];
} buf_send;

vuc_put_meta_t vuc_put_meta[VUC_PC_PUT_ADRS]= {
  { SCOPE_SIZE, scope_in },
};

vuc_get_meta_t vuc_get_meta[VUC_PC_GET_ADRS]= {
};


void vuc_send_buf (buf_send * b);


void vuc_send_8 (u8 data)
{
  if_send_8(data);
}

void vuc_send_8_8 (u8 data1, u8 data2)
{
  if_send_8(data1);
  if_send_8(data2);
}


void print_meta (char * type, u8 len, u8 * data)
{
  char * buf= g_alloca(len+1);

  g_strlcpy(buf, data, len+1);
  dbgf3("meta %s: %s", type, buf);
//  fwrite(data, len, 1, stdout);
//  dbgf3("\n");
}

#if (VUC_PACKET)
int vuc_meta_hs (u8 hs, u8 len, u8 * data)
{
}
#endif

u8 vuc_in_handshake (u8 subcmd, u8 len, u8 * data)
{
  buf_send * b;

  dbgf3(
#if (VUC_PKT_SEQ)
    "meta pkt #%x %s", *data,
#else
    "meta pkt %s",
#endif
    subcmd==VUC_META_PKT_WTF ? "wtf" : subcmd==VUC_META_PKT_OK ? "ok" : "bad"
  );
  if (subcmd==VUC_META_PKT_WTF) {
#if (VUC_PKT_SEQ)
    if (*data!=vuc_in_pkt_hs_no) {
      // Some serious error occured
      // TODO: look into int, reset or do something clever
      // ... but for now just continue with the pkt.no. the uC thinks is right
      vuc_in_pkt_hs_code= VUC_META_PKT_BAD;
      vuc_in_pkt_hs_no= *data;
    }
#endif
    // uC missed last handshake - just repeat it
    vuc_send_8(VUC_CMD_META|vuc_in_pkt_hs_code);
    VUC_OPT(PKT_SEQ, vuc_send_8(vuc_in_pkt_hs_no);)
    return;
  }
  if (subcmd!=VUC_META_PKT_WTF && !vuc_out_pkt_wait)
#if (VUC_PKT_SEQ)
    return errf("redundant handshake %x in #%x.", subcmd, *data);
#else
    return errf("redundant handshake %x.", subcmd);
#endif
  VUC_OPT(PKT_SEQ,
    if (subcmd!=VUC_META_PKT_WTF && *data!=vuc_out_pkt_no)
      return errf("wrong handshake #. Got %x, want %x.", *data, vuc_out_pkt_no);
  )
  if (subcmd==VUC_META_PKT_OK) {
    vuc_out_pkt_wait= 0;
    VUC_OPT(PKT_SEQ, vuc_out_pkt_no++;)
//    dbgf3("Next out-pkt #%x", vuc_out_pkt_no);
    if ((b= g_async_queue_try_pop(send_q)))
      vuc_send_buf(b);
  }
#if (VUC_PKT_SEQ)
  else
    vuc_out_chk_cnt++;
#endif

  return 0;
}

u8 vuc_in_meta (u8 subcmd, u8 len, u8 * data)
{
  if (subcmd==VUC_META_ALIVE)
    { vuc_alive_cnt++; return 0; }

  else if (subcmd==VUC_META_EXT)
    print_meta("ext", len, data);
  else if (subcmd==VUC_META_MSG)
    print_meta("msg", len, data);
  else if (subcmd==VUC_META_DBG)
    { print_meta("dbg", len, data); return 0; }

  else if (subcmd==VUC_META_HELLO)
    dbgf3("meta hello: %d.%d", data[0]>>4, data[0]&0x0f);
  else if (subcmd==VUC_META_ID)
    print_meta("id", len, data);

  else
    print_meta("???", len, data);

  return 1;
}

void vuc_in_err (u8 code, u16 data)
{
  gchar text[60];

  vuc_err_cnt[code]++;

  if (code==VUC_ERR_PKT_HDR || code==VUC_ERR_PKG_BAD) {
    vuc_line_err_cnt++;
  } else {
    vuc_proto_err_cnt++;
  }

  if (code==VUC_ERR_PKT_HDR) {
    dbgf0("err pkt header: %x", data&0xff);
  }
}


#define VUC_IN_DATA_SIZE 256

#define VUC_PUT_ADRS VUC_PC_PUT_ADRS
#define VUC_GET_ADRS VUC_PC_GET_ADRS
u8 vuc_out_on= 1;
#include "vuc_input.h"


// a packet-header just came in
void vuc_in_pkt_start()
{
//  pkt_start= my_time();
}

// send handshake to uC
void vuc_in_send_hs (i8 rep)
{
  if (rep) {
    vuc_in_pkt_hs_code= rep<0 ? VUC_META_PKT_BAD : VUC_META_PKT_OK;
    vuc_in_pkt_hs_no= vuc_in_pkt_no;
    vuc_send_8(VUC_CMD_META|vuc_in_pkt_hs_code);
    VUC_OPT(PKT_SEQ, vuc_send_8(vuc_in_pkt_hs_no);)
#if (VUC_DBG_OUT)
#  if (VUC_PKT_SEQ)
    dbgf2("OUT: %02x %02x", VUC_CMD_META|vuc_in_pkt_hs_code, vuc_in_pkt_no);
#  else
    dbgf2("OUT: %02x", VUC_CMD_META|vuc_in_pkt_hs_code);
#  endif
#endif
#if (VUC_PKT_CHK)
    if (rep<0)
      vuc_in_chk_cnt++;
#endif
  }
}


void vuc_send_buf (buf_send * b)
{
#if (VUC_PACKET)
  gint i, j= 0;
#if (VUC_PKT_CHK)
  u8 chk= 0;
#endif
#if (VUC_DBG_OUT)
  GString * dbg_hex= g_string_new(NULL);
#endif

  b->pkt= 1;
  VUC_OPT(PACKET,
    vuc_out_pkt_wait= 1;
    b->buf[j++]= VUC_CMD_META|VUC_META_PKT;
    b->buf[j++]= b->end-b->start;
  )
  VUC_OPT(PKT_SEQ, b->buf[j++]= vuc_out_pkt_no;)
#if (VUC_PKT_CHK)
  for (i=b->start; i<b->end; i++)
    chk+= b->buf[i];
  b->buf[b->end++]= chk;
#endif
  b->start= 0;
#endif

  for (i=b->start; i<b->end; i++) {
    if_send_8(b->buf[i]);
#if (VUC_DBG_OUT)
    g_string_append_printf(dbg_hex, " %02x", b->buf[i]);
#endif
  }
#if (VUC_DBG_OUT)
  dbgf2("OUT:%s", dbg_hex->str); g_string_free(dbg_hex, 1);
#endif
}

void vuc_send_pkt (buf_send * b)
{
  if (vuc_out_pkt_wait)
    g_async_queue_push(send_q, b);
  else
    vuc_send_buf(b);
}

/* Send some 8-bit-values to uC */
void vuc_put_8 (guint channel, guint count, ...)
{
  va_list vals;
  gint v, i, j= 2;
  buf_send * b= g_malloc(sizeof(buf_send)+4+1+count);

  VUC_OPT(PKT_SEQ, j++;)
  b->pkt= 0; b->start= j; b->end= j+1+count;
#if (VUC_DBG_OUT)
  GString * dbg_hex= g_string_new(NULL);
#endif

  // TODO: channel-numbers above 64
  b->buf[j++]= VUC_CMD_PUT|channel;
#if (VUC_DBG_OUT)
  g_string_append_printf(dbg_hex, "%02x =", VUC_CMD_PUT|channel);
#endif
  va_start(vals, count);
  for (i=0; i<count; i++) {
    v= va_arg(vals, int);
    b->buf[j++]= v;
#if (VUC_DBG_OUT)
    g_string_append_printf(dbg_hex, " %02x", (u8)v);
#endif
  }
  va_end(vals);

#if (VUC_DBG_OUT)
  dbgf3("PUT: %s", dbg_hex->str); g_string_free(dbg_hex, 1);
#endif
  vuc_send_pkt(b);
}

/* Send some 16-bit-values to uC */
void vuc_put_16 (guint channel, guint count, ...)
{
  va_list vals;
  gint v, i, j= 2;
  buf_send * b= g_malloc(sizeof(buf_send)+4+1+count*2);

  VUC_OPT(PKT_SEQ, j++;)
  b->pkt= FALSE; b->start= j; b->end= j+1+count*2;
#if (VUC_DBG_OUT)
  GString * dbg_hex= g_string_new(NULL);
#endif

  // TODO: channel-numbers above 64
  b->buf[j++]= VUC_CMD_PUT|channel;
#if (VUC_DBG_OUT)
  g_string_append_printf(dbg_hex, "%02x =", VUC_CMD_PUT|channel);
#endif
  va_start(vals, count);
  for (i=0; i<count; i++) {
    v= va_arg(vals, int);
    b->buf[j++]= v&0xff;
    b->buf[j++]= v>>8;
#if (VUC_DBG_OUT)
    g_string_append_printf(dbg_hex, " %04x", v);
#endif
  }
  va_end(vals);

#if (VUC_DBG_OUT)
  dbgf3("PUT: %s", dbg_hex->str); g_string_free(dbg_hex, 1);
#endif
  vuc_send_pkt(b);
}

#define IO_IN_BUF_SIZE 1024

gint vuc_io_poll ()
{
  gint size, read= 0;
  gchar io_in_buf[IO_IN_BUF_SIZE];
  guchar c;
#if (VUC_DBG_IN)
  GString * dbg_hex= g_string_new(NULL);
  GString * dbg_asc= g_string_new(NULL);
#endif

//  GIOStatus
  size= if_receive(io_in_buf, IO_IN_BUF_SIZE);
  if (size<=0)
    return;
  while (read<size) {
    c= io_in_buf[read];
    vuc_input(io_in_buf[read++]);
#if (VUC_DBG_IN)
    g_string_append_printf(dbg_hex, " %02x", c);
    g_string_append_printf(dbg_asc, "%c", (c<0x20 || c>=0x80) ? '.' : c);
#endif
  }
#if (VUC_DBG_IN)
  dbgf1("IN: %s  \t%s", dbg_hex->str, dbg_asc->str);
  g_string_free(dbg_hex, 1); g_string_free(dbg_asc, 1);
#endif
}

gboolean vuc_io_timeout (gpointer data)
{
  vuc_io_poll();
  return 1;
}


/* Send "meta"-cmd w/o parameter */
void vuc_meta_0 (u8 meta)
{
  meta|= VUC_CMD_META;
  if_send_8(meta);

#if (VUC_DBG_OUT)
  dbgf2("OUT: %02x", meta);
#endif
}



/*** GUI *********************************************************************/

void on_window1_destroy (GtkWidget * widget, gpointer udata)
{
  gtk_main_quit();
}


/* TODO: have this auto-generated ********************************************/

GtkToggleButton * u_gen_power;
GtkAdjustment * u_gen_freq;
GtkAdjustment * u_gen_pw;

GtkDrawingArea * u_scope_area;
GtkToggleButton * u_scope_tb[8];


/*** Gen *********************************************************************/

gdouble u_gen_freq_v= 20;
gdouble u_gen_pw_v= 1;

void u_gen_set ()
{
  u16 f= (u16)(F_CPU/8/2/u_gen_freq_v);
  u16 p= (u16)(f*(u_gen_pw_v/100));
  if (p==0)
    p++;
  if (p==f)
    p--;
  vuc_put_16(0, 2, f, p);
}

void on_u_gen_power_toggled (GtkWidget * widget, gpointer udata)
{
  u8 v= gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  dbgf4("Power: %d", v);
  vuc_put_8(1, 1, v);
}

void on_u_gen_freq_value_changed (GtkWidget * widget, gpointer udata)
{
  u_gen_freq_v= gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));
#if (DBG_GUI)
  dbgf4("Freq: %04.0f", u_gen_freq_v);
#endif
  u_gen_set();
}

void on_u_gen_pw_value_changed (GtkWidget * widget, gpointer udata)
{
  u_gen_pw_v= gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));
#if (DBG_GUI)
  dbgf4("PW: %02.0f", u_gen_pw_v);
#endif
  u_gen_set();
}

/*** Scope *******************************************************************/

#define SC_NORM 1024

// Timebase constants in uS/div
gint sc_timebase[8]= {
  500, 1000, 2000, 5000, 10000, 20000, 50000, 100000
};

gint sc_count= 0;
GdkPoint * sc_points= NULL;

GdkPixmap * pixmap= NULL;
GdkGC * gc= NULL;
gint width= 0;
gint height= 0;

void draw_pixmap ()
{
  GdkColormap * colormap= NULL;
  GdkColor fg;
  gint i, o, x, ox, oy, mx, my;
  GdkPoint * points= g_alloca(sizeof(GdkPoint)*max(sc_count, 2));

  if (pixmap) {
    gdk_gc_unref(gc);
    gdk_pixmap_unref(pixmap);
  }
  pixmap= gdk_pixmap_new(GTK_WIDGET(u_scope_area)->window, width, height, -1);
  gdk_draw_rectangle(pixmap, GTK_WIDGET(u_scope_area)->style->black_gc, TRUE, 0, 0, width, height);
  gc= gdk_gc_new(pixmap);
  colormap= gdk_gc_get_colormap(gc);

  ox= width*5/100;  mx= width-(2*ox);
  oy= height*5/100; my= height-(2*oy);

  if (!gdk_color_parse(SCOPE_DIV_COLOR, &fg))
    gdk_color_white(colormap, &fg);
  gdk_colormap_alloc_color(colormap, &fg, TRUE, TRUE);
  gdk_gc_set_foreground(gc, &fg);
  for (i=0; i<=8; i++) {
    o= i==4 ? ox/2 : ox;
    points[0].x= o;        points[0].y= oy+i*my/8;
    points[1].x= width-o;  points[1].y= oy+i*my/8;
    gdk_draw_lines(pixmap, gc, points, 2);
  }
  for (i=0; i<=8; i++) {
    o= i==4 ? oy/2 : oy;
    points[0].x= ox+i*mx/8; points[0].y= o;
    points[1].x= ox+i*mx/8; points[1].y= height-o;
    gdk_draw_lines(pixmap, gc, points, 2);
  }

  if (!gdk_color_parse(SCOPE_COLOR, &fg))
    gdk_color_white(colormap, &fg);
  gdk_colormap_alloc_color(colormap, &fg, TRUE, TRUE);
  gdk_gc_set_foreground(gc, &fg);
  for (x=0; x<sc_count; x++) {
    points[x].x= ox+sc_points[x].x*mx/SC_NORM;
    points[x].y= oy+sc_points[x].y*my/SC_NORM;
  }
  gdk_draw_lines(pixmap, gc, points, sc_count);
}

void draw_screen (gint x, gint y, gint w, gint h)
{
  GdkDrawable * drawable= GTK_WIDGET(u_scope_area)->window;

  gdk_draw_drawable(drawable, gc, pixmap,  x, y,  x, y,  w, h);
}

i8 scope_in (u8 * data)
{
  u8 x, v;
  gint c, lc= 255-data[0];

//  dbgf("scope_in()");

//  vuc_err(0, 0);

#if (VUC_TEST_PROTO==1)
  static u8 diag_cnt= 0;
  gint fl= 0;

  for (x=0; x<SCOPE_SIZE; x++) {
    v= x ? x : diag_cnt++;
    c= data[x];
    if (c!=v) {
      printf(" !%02x:%02x ", v, c);
      fl= 1;
      if (!x)
        diag_cnt= c+1;
    }
  }
  if (fl) {
    scope_diag_err++;
    printf("\n");
//    vuc_meta_0(VUC_META_NAK);
//    vuc_err(VUC_ERR__MAX+1, 0);
  } else
    scope_diag_ok++;

  return 1;
#endif

  sc_count= 0;
  for (x=0; x<SCOPE_SIZE; x++) {
    c= 255-data[x];
    if (x==0 || c!=lc) {
      sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= lc*SC_NORM/256; sc_count++;
      sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= c*SC_NORM/256; sc_count++;
    }
    lc= c;
  }
  sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= c*SC_NORM/256; sc_count++;
  draw_pixmap();
  draw_screen(0, 0, width, height);
  return 1;
}

gboolean on_u_scope_area_configure_event (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
  width= widget->allocation.width;
  height= widget->allocation.height;

  draw_pixmap();
  return TRUE;
}

gboolean on_u_scope_area_expose_event (GtkWidget * widget, GdkEventExpose * event, gpointer user_data)
{
  draw_screen(event->area.x, event->area.y, event->area.width, event->area.height);
  return FALSE;
}

gboolean in_scope_tb= 0;

gboolean on_u_scope_tb_toggled (GtkWidget * widget, gpointer udata)
{
  gint i, t, v= (int)udata;

  if (in_scope_tb)
    return FALSE;

  in_scope_tb= TRUE;
  for (i=0; i<8; i++) {
    gtk_toggle_button_set_active(u_scope_tb[i], i==v);
  }
  in_scope_tb= FALSE;

  t= (F_CPU/SCOPE_PRESC_VAL) / (1000000/(sc_timebase[v]*8/SCOPE_SIZE));
  dbgf("Timebase: %d -> %5.1f mS/div -> t=%d", (int)udata, sc_timebase[v]/1000.0, t);
  vuc_put_8(2, 1, t);

  return FALSE;
}

/* /TODO *********************************************************************/

void show_counter (gint * cnt, gint * last, gint max, gchar * fmt, GtkEntry * entry)
{
  gchar text[10];

  if (entry && *last!=*cnt) {
    if (*cnt>max)
      *cnt= 0;
    snprintf(text, 10, fmt, *cnt);
    gtk_entry_set_text(entry, text);
    *last= *cnt;
  }
}


gboolean backgr_func (gpointer data)
{
  static gint last_alive= 0;
  static gint last_line_err= 0;
  static gint last_proto_err= 0;
  static gint last_scope_diag_ok= 0;
  static gint last_scope_diag_err= 0;
  gchar text[32];

  show_counter(&vuc_alive_cnt, &last_alive, 999, "%03d", gui_alive_cnt);
#if (VUC_PACKET)
  show_counter(&vuc_in_chk_cnt, &last_line_err, 999, "%03d", gui_in_chk_cnt);
  show_counter(&vuc_out_chk_cnt, &last_line_err, 999, "%03d", gui_out_chk_cnt);
  show_counter(&vuc_line_err_cnt, &last_line_err, 9999, "%04d", gui_line_err_cnt);
#endif
  show_counter(&vuc_proto_err_cnt, &last_proto_err, 9999, "%04d", gui_proto_err_cnt);

  if (last_scope_diag_ok!=scope_diag_ok || last_scope_diag_err!=scope_diag_err) {
    last_scope_diag_ok= scope_diag_ok; last_scope_diag_err= scope_diag_err;
    snprintf(text, 32, "OK: %06d  ERR: %06d", scope_diag_ok, scope_diag_err);
    gtk_entry_set_text(gui_status_msg, text);
  }

  gpointer p= g_async_queue_try_pop(console_q);
  if (p) {
    fprintf(stderr, "%s\n", (char *)p);
    g_free(p);
  }

  return 1;
}

GTimeVal start_time;

int main (int argc, char * argv[])
{
  gint i;
  GError * error;
  GtkBuilder * builder;
  GtkWindow * window1;
  GtkPaned * hpaned1;

  GtkScrolledWindow * dbg_scr;
  GtkAdjustment * dbg_adj;
  GtkTextView * dbg_buf;

  GThread * send_th;

  GtkStyle * style;


  g_get_current_time(&start);

  if (if_init(argc, argv)>1)
    exit(1);

  gtk_init(&argc, &argv);

  builder= gtk_builder_new();
  if (!gtk_builder_add_from_file(builder, glade, NULL)) {
    printf("Need input!\n");
    exit(2);
  }

  memset(vuc_err_cnt, 0, sizeof(gint)*256);

  send_q= g_async_queue_new();
  console_q= g_async_queue_new();

  window1= GTK_WINDOW(gtk_builder_get_object(builder, "window1"));
  gtk_signal_connect(GTK_OBJECT(window1), "destroy", GTK_SIGNAL_FUNC(on_window1_destroy), NULL);

  hpaned1= GTK_PANED(gtk_builder_get_object(builder, "hpaned1"));

  gui_alive_cnt= GTK_ENTRY(gtk_builder_get_object(builder, "alive_cnt"));
  gui_in_chk_cnt= GTK_ENTRY(gtk_builder_get_object(builder, "in_chk_cnt"));
  gui_out_chk_cnt= GTK_ENTRY(gtk_builder_get_object(builder, "out_chk_cnt"));
  gui_line_err_cnt= GTK_ENTRY(gtk_builder_get_object(builder, "line_err_cnt"));
  gui_proto_err_cnt= GTK_ENTRY(gtk_builder_get_object(builder, "proto_err_cnt"));
  gui_status_msg= GTK_ENTRY(gtk_builder_get_object(builder, "status_msg"));

  style= gtk_style_copy(gtk_widget_get_style(GTK_WIDGET(gui_alive_cnt)));
  memcpy(&style->text[GTK_STATE_INSENSITIVE], &style->text[GTK_STATE_NORMAL], sizeof(GdkColor));
  memcpy(&style->base[GTK_STATE_INSENSITIVE], &style->base[GTK_STATE_NORMAL], sizeof(GdkColor));

  gtk_widget_set_style(GTK_WIDGET(gui_alive_cnt), style);
  gtk_widget_set_style(GTK_WIDGET(gui_in_chk_cnt), style);
  gtk_widget_set_style(GTK_WIDGET(gui_out_chk_cnt), style);
  gtk_widget_set_style(GTK_WIDGET(gui_line_err_cnt), style);
  gtk_widget_set_style(GTK_WIDGET(gui_proto_err_cnt), style);
  gtk_widget_set_style(GTK_WIDGET(gui_status_msg), style);

  dbg_scr= GTK_SCROLLED_WINDOW(gtk_builder_get_object(builder, "scrolledwindow1"));
  if (dbg_scr) {
    dbg_buf= GTK_TEXT_VIEW(gtk_builder_get_object(builder, "textview1"));
  }


  /* TODO: have this auto-generated ******************************************/
  u_gen_power= GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "u_gen_power"));
  gtk_signal_connect(GTK_OBJECT(u_gen_power), "toggled", GTK_SIGNAL_FUNC(on_u_gen_power_toggled), NULL);

  u_gen_freq= GTK_ADJUSTMENT(gtk_builder_get_object(builder, "u_gen_freq"));
  gtk_adjustment_set_value(u_gen_freq, 0);
  gtk_signal_connect(GTK_OBJECT(u_gen_freq), "value-changed", GTK_SIGNAL_FUNC(on_u_gen_freq_value_changed), NULL);

  u_gen_pw= GTK_ADJUSTMENT(gtk_builder_get_object(builder, "u_gen_pw"));
  gtk_adjustment_set_value(u_gen_pw, 0);
  gtk_signal_connect(GTK_OBJECT(u_gen_pw), "value-changed", GTK_SIGNAL_FUNC(on_u_gen_pw_value_changed), NULL);

  sc_points= g_malloc0(sizeof(GdkPoint)*(SCOPE_SIZE+1)*2);
  sc_points[0].x= 0;       sc_points[0].y= SC_NORM/2;
  sc_points[1].x= SC_NORM; sc_points[1].y= SC_NORM/2;
  sc_count= 2;
  u_scope_area= GTK_DRAWING_AREA(gtk_builder_get_object(builder, "u_scope_area"));
  gtk_widget_add_events(GTK_WIDGET(u_scope_area), GDK_EXPOSURE_MASK|GDK_LEAVE_NOTIFY_MASK|GDK_BUTTON_PRESS_MASK|GDK_POINTER_MOTION_MASK|GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect(GTK_OBJECT(u_scope_area), "configure-event", GTK_SIGNAL_FUNC(on_u_scope_area_configure_event), NULL);
  gtk_signal_connect(GTK_OBJECT(u_scope_area), "expose-event", GTK_SIGNAL_FUNC(on_u_scope_area_expose_event), NULL);

  for (i=0; i<8; i++) {
    char buf[20];
    sprintf(buf, "u_scope_tb_%d", i+1);
    u_scope_tb[i]= GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, buf));
    gtk_signal_connect(GTK_OBJECT(u_scope_tb[i]), "toggled", GTK_SIGNAL_FUNC(on_u_scope_tb_toggled), (void *)i);
  }
  /* /TODO *******************************************************************/


  g_timeout_add(10, backgr_func, NULL);

  gtk_widget_show(GTK_WIDGET(window1));

  gtk_main();

  return 0;
}
